@@ -151,10 +151,6 @@ xsltFreeCompMatch(xsltCompMatchPtr comp) {
 	return;
     if (comp->pattern != NULL)
 	xmlFree((xmlChar *)comp->pattern);
-    if (comp->mode != NULL)
-	xmlFree((xmlChar *)comp->mode);
-    if (comp->modeURI != NULL)
-	xmlFree((xmlChar *)comp->modeURI);
     if (comp->nsList != NULL)
 	xmlFree(comp->nsList);
     for (i = 0;i < comp->nbStep;i++) {
@@ -373,7 +369,7 @@ xsltReverseCompMatch(xsltCompMatchPtr comp) {
  * @mode:  the mode name or NULL
  * @modeURI:  the mode URI or NULL
  *
- * Test wether the node matches the pattern
+ * Test whether the node matches the pattern
  *
  * Returns 1 if it matches, 0 if it doesn't and -1 in case of failure
  */
@@ -392,7 +388,10 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
     if (mode != NULL) {
 	if (comp->mode == NULL)
 	    return(0);
-	if ((comp->mode != mode) && (!xmlStrEqual(comp->mode, mode)))
+	/*
+	 * both mode strings must be interned on the stylesheet dictionary
+	 */
+	if (comp->mode != mode)
 	    return(0);
     } else {
 	if (comp->mode != NULL)
@@ -401,8 +400,10 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
     if (modeURI != NULL) {
 	if (comp->modeURI == NULL)
 	    return(0);
-	if ((comp->modeURI != modeURI) &&
-	    (!xmlStrEqual(comp->modeURI, modeURI)))
+	/*
+	 * both modeURI strings must be interned on the stylesheet dictionary
+	 */
+	if (comp->modeURI != modeURI)
 	    return(0);
     } else {
 	if (comp->modeURI != NULL)
@@ -521,14 +522,16 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
             case XSLT_OP_ANCESTOR:
 		/* TODO: implement coalescing of ANCESTOR/NODE ops */
 		if (step->value == NULL) {
-		    i++;
-		    step = &comp->steps[i];
+		    step = &comp->steps[i+1];
 		    if (step->op == XSLT_OP_ROOT)
 			return(1);
-		    if (step->op != XSLT_OP_ELEM)
+		    /* added NS, ID and KEY as a result of bug 168208 */
+		    if ((step->op != XSLT_OP_ELEM) && 
+				(step->op != XSLT_OP_ALL) && 
+				(step->op != XSLT_OP_NS) &&
+				(step->op != XSLT_OP_ID) &&
+				(step->op != XSLT_OP_KEY))
 			return(0);
-		    if (step->value == NULL)
-			return(-1);
 		}
 		if (node == NULL)
 		    return(0);
@@ -540,6 +543,11 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    (node->type == XML_NAMESPACE_DECL))
 		    return(0);
 		node = node->parent;
+		if ((step->op != XSLT_OP_ELEM) && step->op != XSLT_OP_ALL)
+			continue;
+		i++;
+		if (step->value == NULL)
+		    continue;
 		while (node != NULL) {
 		    if (node == NULL)
 			return(0);
@@ -607,8 +615,20 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		break;
 	    case XSLT_OP_PREDICATE: {
 		xmlNodePtr oldNode;
+		xmlDocPtr doc;
 		int oldCS, oldCP;
 		int pos = 0, len = 0;
+		int isRVT;
+
+		doc = node->doc;
+		if ((doc != NULL) &&
+		    (doc->name != NULL) &&
+		    (doc->name[0] == ' ') &&
+		    (xmlStrEqual(BAD_CAST doc->name,
+		    		 BAD_CAST " fake node libxslt")))
+		    isRVT = 1;
+		else
+		    isRVT = 0;
 		/*
 		 * The simple existing predicate code cannot handle
 		 * properly cascaded predicates. If in this situation
@@ -616,19 +636,17 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		 * if the node is in the result list.
 		 */
 		if (comp->steps[i + 1].op == XSLT_OP_PREDICATE) {
-		    xmlDocPtr prevdoc, doc;
+		    xmlDocPtr prevdoc;
 		    xmlXPathObjectPtr list;
 		    int ix, j;
 		    int nocache = 0;
 
 		    prevdoc = (xmlDocPtr)
-			XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra);
-		    ix = (int)
-			XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra);
+			XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra, ptr);
+		    ix = XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra, ival);
 		    list = (xmlXPathObjectPtr)
 			XSLT_RUNTIME_EXTRA_LST(ctxt, sel->lenExtra);
 		    
-		    doc = node->doc;
 		    if ((list == NULL) || (prevdoc != doc)) {
 			xmlChar *query;
 			xmlXPathObjectPtr newlist;
@@ -658,15 +676,8 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			}
 			ix = 0;
 
-			if ((parent == NULL) || (node->doc == NULL))
+			if ((parent == NULL) || (node->doc == NULL) || isRVT)
 			    nocache = 1;
-			else {
-			    if ((doc->name != NULL) &&
-				(doc->name[0] == ' ') &&
-				(xmlStrEqual(BAD_CAST doc->name, 
-					     BAD_CAST " fake node libxslt")))
-				nocache = 1;
-			}
 			
 			if (nocache == 0) {
 			    if (list != NULL)
@@ -675,9 +686,9 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 
 			    XSLT_RUNTIME_EXTRA_LST(ctxt, sel->lenExtra) =
 				(void *) list;
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra) =
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra, ptr) =
 				(void *) doc;
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra) =
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra, ival) =
 			        0;
 			    XSLT_RUNTIME_EXTRA_FREE(ctxt, sel->lenExtra) =
 				(xmlFreeFunc) xmlXPathFreeObject;
@@ -722,9 +733,8 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    int ix, nocache = 0;
 
 		    previous = (xmlNodePtr)
-			XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra);
-		    ix = (int)
-			XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra);
+			XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra, ptr);
+		    ix = XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra, ival);
 		    if ((previous != NULL) &&
 			(previous->parent == node->parent)) {
 			/*
@@ -767,18 +777,19 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			if (sibling != NULL) {
 			    pos = ix + indx;
 			    /*
-			     * If the node is in a Value Tree we cannot
-			     * cache it !
+			     * If the node is in a Value Tree we need to
+			     * save len, but cannot cache the node!
+			     * (bugs 153137 and 158840)
 			     */
-			    if ((node->doc != NULL) &&
-			        (node->doc->name != NULL) &&
-				(node->doc->name[0] != ' ')) {
-				len = (int)
-				    XSLT_RUNTIME_EXTRA(ctxt, sel->lenExtra);
-				XSLT_RUNTIME_EXTRA(ctxt,
-					sel->previousExtra) = node;
-				XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra) =
-				    (void *) pos;
+			    if (node->doc != NULL) {
+				len = XSLT_RUNTIME_EXTRA(ctxt,
+				        sel->lenExtra, ival);
+				if (!isRVT) {
+				    XSLT_RUNTIME_EXTRA(ctxt,
+					sel->previousExtra, ptr) = node;
+				    XSLT_RUNTIME_EXTRA(ctxt,
+				        sel->indexExtra, ival) = pos;
+				}
 			    }
 			    ix = pos;
 			} else
@@ -826,13 +837,14 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			 * If the node is in a Value Tree we cannot
 			 * cache it !
 			 */
-			if ((node->doc != NULL) && (nocache == 0)) {
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra) =
+			if ((!isRVT) && (node->doc != NULL) &&
+			    (nocache == 0)) {
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra, ptr) =
 				node;
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra) =
-				(void *) pos;
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->lenExtra) =
-				(void *) len;
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra, ival) =
+				pos;
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->lenExtra, ival) =
+				len;
 			}
 		    }
 		} else if ((sel != NULL) && (sel->op == XSLT_OP_ALL) &&
@@ -841,9 +853,8 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    int ix, nocache = 0;
 
 		    previous = (xmlNodePtr)
-			XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra);
-		    ix = (int)
-			XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra);
+			XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra, ptr);
+		    ix = XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra, ival);
 		    if ((previous != NULL) &&
 			(previous->parent == node->parent)) {
 			/*
@@ -877,15 +888,13 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			     * If the node is in a Value Tree we cannot
 			     * cache it !
 			     */
-			    if ((node->doc != NULL) &&
-			        (node->doc->name != NULL) &&
-				(node->doc->name[0] != ' ')) {
-				len = (int)
-				    XSLT_RUNTIME_EXTRA(ctxt, sel->lenExtra);
+			    if ((node->doc != NULL) && !isRVT) {
+				len = XSLT_RUNTIME_EXTRA(ctxt,
+				        sel->lenExtra, ival);
 				XSLT_RUNTIME_EXTRA(ctxt,
-					sel->previousExtra) = node;
-				XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra) =
-				    (void *) pos;
+					sel->previousExtra, ptr) = node;
+				XSLT_RUNTIME_EXTRA(ctxt,
+					sel->indexExtra, ival) = pos;
 			    }
 			} else
 			    pos = 0;
@@ -923,13 +932,13 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			 * If the node is in a Value Tree we cannot
 			 * cache it !
 			 */
-			if ((node->doc != NULL) && (nocache == 0)) {
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra) =
+			if ((node->doc != NULL) && (nocache == 0) && !isRVT) {
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->previousExtra, ptr) =
 				node;
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra) =
-				(void *) pos;
-			    XSLT_RUNTIME_EXTRA(ctxt, sel->lenExtra) =
-				(void *) len;
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->indexExtra, ival) =
+				pos;
+			    XSLT_RUNTIME_EXTRA(ctxt, sel->lenExtra, ival) =
+				len;
 			}
 		    }
 		}
@@ -999,7 +1008,7 @@ wrong_index:
  * @node: a node
  * @comp: the precompiled pattern list
  *
- * Test wether the node matches one of the patterns in the list
+ * Test whether the node matches one of the patterns in the list
  *
  * Returns 1 if it matches, 0 if it doesn't and -1 in case of failure
  */
@@ -1612,7 +1621,7 @@ xsltCompileRelativePathPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 	    PUSH(XSLT_OP_PARENT, NULL, NULL);
 	    NEXT;
 	    SKIP_BLANKS;
-	    if ((CUR != 0) || (CUR == '|')) {
+	    if ((CUR != 0) && (CUR != '|')) {
 		xsltCompileRelativePathPattern(ctxt, NULL);
 	    }
 	} else {
@@ -1656,7 +1665,7 @@ xsltCompileLocationPathPattern(xsltParserContextPtr ctxt) {
 	NEXT;
 	SKIP_BLANKS;
 	PUSH(XSLT_OP_ROOT, NULL, NULL);
-	if ((CUR != 0) || (CUR == '|')) {
+	if ((CUR != 0) && (CUR != '|')) {
 	    PUSH(XSLT_OP_PARENT, NULL, NULL);
 	    xsltCompileRelativePathPattern(ctxt, NULL);
 	}
@@ -1905,9 +1914,9 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	
 	pat->template = cur;
 	if (mode != NULL)
-	    pat->mode = xmlStrdup(mode);
+	    pat->mode = xmlDictLookup(style->dict, mode, -1);
 	if (modeURI != NULL)
-	    pat->modeURI = xmlStrdup(modeURI);
+	    pat->modeURI = xmlDictLookup(style->dict, modeURI, -1);
 	if (priority != XSLT_PAT_NO_PRIORITY)
 	    pat->priority = priority;
 
@@ -2112,7 +2121,7 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	}
 	if (name != NULL) {
 	    /*
-	     * find the list of appliable expressions based on the name
+	     * find the list of applicable expressions based on the name
 	     */
 	    list = (xsltCompMatchPtr) xmlHashLookup3(curstyle->templatesHash,
 					     name, ctxt->mode, ctxt->modeURI);
